"""Data class to hold common runtime arguments for dataset creation."""

import re
from dataclasses import dataclass
from importlib.metadata import version

from hipscat.io import file_io

# pylint: disable=too-many-instance-attributes


@dataclass
class RuntimeArguments:
    """Data class for holding runtime arguments"""

    ## Output
    output_path: str = ""
    """base path where new catalog should be output"""
    output_catalog_name: str = ""
    """short, convenient name for the catalog"""

    ## Execution
    tmp_dir: str = ""
    """path for storing intermediate files"""
    overwrite: bool = False
    """if there is existing data at the `catalog_path`, 
    should we overwrite and create a new catalog"""
    progress_bar: bool = True
    """if true, a tqdm progress bar will be displayed for user
    feedback of map reduce progress"""
    dask_tmp: str = ""
    """ directory for dask worker space. this should be local to
    the execution of the pipeline, for speed of reads and writes"""
    dask_n_workers: int = 1
    """number of workers for the dask client"""
    dask_threads_per_worker: int = 1
    """number of threads per dask worker"""

    completion_email_address: str = ""
    """if provided, send an email to the indicated email address once the 
    import pipeline has complete."""

    catalog_path = ""
    """constructed output path for the catalog that will be something like
    <output_path>/<output_catalog_name>"""
    tmp_path = ""
    """constructed temp path - defaults to tmp_dir, then dask_tmp, but will create
    a new temp directory under catalog_path if no other options are provided"""

    def __post_init__(self):
        self._check_arguments()

    def _check_arguments(self):
        if not self.output_path:
            raise ValueError("output_path is required")
        if not self.output_catalog_name:
            raise ValueError("output_catalog_name is required")
        if re.search(r"[^A-Za-z0-9_\-\\]", self.output_catalog_name):
            raise ValueError("output_catalog_name contains invalid characters")

        if self.dask_n_workers <= 0:
            raise ValueError("dask_n_workers should be greather than 0")
        if self.dask_threads_per_worker <= 0:
            raise ValueError("dask_threads_per_worker should be greater than 0")

        if not file_io.does_file_or_directory_exist(self.output_path):
            raise FileNotFoundError(
                f"output_path ({self.output_path}) not found on local storage"
            )

        self.catalog_path = file_io.append_paths_to_pointer(
            self.output_path, self.output_catalog_name
        )
        if not self.overwrite:
            if file_io.directory_has_contents(self.catalog_path):
                raise ValueError(
                    f"output_path ({self.catalog_path}) contains files."
                    " choose a different directory or use --overwrite flag"
                )
        file_io.make_directory(self.catalog_path, exist_ok=True)

        if self.tmp_dir:
            if not file_io.does_file_or_directory_exist(self.tmp_dir):
                raise FileNotFoundError(
                    f"tmp_dir ({self.tmp_dir}) not found on local storage"
                )
            self.tmp_path = file_io.append_paths_to_pointer(
                self.tmp_dir, self.output_catalog_name, "intermediate"
            )
        elif self.dask_tmp:
            if not file_io.does_file_or_directory_exist(self.dask_tmp):
                raise FileNotFoundError(
                    f"dask_tmp ({self.dask_tmp}) not found on local storage"
                )
            self.tmp_path = file_io.append_paths_to_pointer(
                self.dask_tmp, self.output_catalog_name, "intermediate"
            )
        else:
            self.tmp_path = file_io.append_paths_to_pointer(
                self.catalog_path, "intermediate"
            )
        file_io.make_directory(self.tmp_path, exist_ok=True)

    def provenance_info(self) -> dict:
        """Fill all known information in a dictionary for provenance tracking.

        Returns:
            dictionary with all argument_name -> argument_value as key -> value pairs.
        """
        runtime_args = {
            "catalog_name": self.output_catalog_name,
            "output_path": str(self.output_path),
            "output_catalog_name": self.output_catalog_name,
            "tmp_dir": str(self.tmp_dir),
            "overwrite": self.overwrite,
            "dask_tmp": str(self.dask_tmp),
            "dask_n_workers": self.dask_n_workers,
            "dask_threads_per_worker": self.dask_threads_per_worker,
            "catalog_path": self.catalog_path,
            "tmp_path": str(self.tmp_path),
        }

        runtime_args.update(self.additional_runtime_provenance_info())
        provenance_info = {
            "tool_name": "hipscat_import",
            "version": version("hipscat-import"),
            "runtime_args": runtime_args,
        }

        return provenance_info

    def additional_runtime_provenance_info(self):
        """Any additional runtime args to be included in provenance info from subclasses"""
        return {}



"""Utility to hold all arguments required throughout indexing"""

from dataclasses import dataclass, field
from typing import List, Optional

from hipscat.catalog import Catalog
from hipscat.catalog.index.index_catalog_info import IndexCatalogInfo

from hipscat_import.runtime_arguments import RuntimeArguments


@dataclass
class IndexArguments(RuntimeArguments):
    """Data class for holding indexing arguments"""

    ## Input
    input_catalog_path: str = ""
    input_catalog: Optional[Catalog] = None
    indexing_column: str = ""
    extra_columns: List[str] = field(default_factory=list)

    ## Output
    include_hipscat_index: bool = True
    include_order_pixel: bool = True

    compute_partition_size: int = 1_000_000_000

    def __post_init__(self):
        self._check_arguments()

    def _check_arguments(self):
        super()._check_arguments()
        if not self.input_catalog_path:
            raise ValueError("input_catalog_path is required")
        if not self.indexing_column:
            raise ValueError("indexing_column is required")

        if not self.include_hipscat_index and not self.include_order_pixel:
            raise ValueError(
                "At least one of include_hipscat_index or include_order_pixel must be True"
            )

        self.input_catalog = Catalog.read_from_hipscat(
            catalog_path=self.input_catalog_path
        )

        if self.compute_partition_size < 100_000:
            raise ValueError("compute_partition_size must be at least 100_000")

    def to_catalog_info(self, total_rows) -> IndexCatalogInfo:
        """Catalog-type-specific dataset info."""
        info = {
            "catalog_name": self.output_catalog_name,
            "total_rows": total_rows,
            "catalog_type": "index",
            "primary_catalog": str(self.input_catalog_path),
            "indexing_column": self.indexing_column,
            "extra_columns": self.extra_columns,
        }
        return IndexCatalogInfo(**info)

    def additional_runtime_provenance_info(self) -> dict:
        return {
            "input_catalog_path": str(self.input_catalog_path),
            "indexing_column": self.indexing_column,
            "extra_columns": self.extra_columns,
            "include_hipscat_index": str(self.include_hipscat_index),
            "include_order_pixel": self.include_order_pixel,
        }



"""Functional tests for catalog import"""

import os
import shutil

import pytest
from hipscat.catalog.catalog import Catalog

import hipscat_import.catalog.resume_files as rf
import hipscat_import.catalog.run_import as runner
from hipscat_import.catalog.arguments import ImportArguments


def test_empty_args():
    """Runner should fail with empty arguments"""
    with pytest.raises(ValueError, match="args is required"):
        runner.run(None, None)


def test_bad_args():
    """Runner should fail with mis-typed arguments"""
    args = {"output_catalog_name": "bad_arg_type"}
    with pytest.raises(ValueError, match="ImportArguments"):
        runner.run(args, None)


@pytest.mark.dask
def test_resume_dask_runner(
    dask_client,
    small_sky_parts_dir,
    resume_dir,
    tmp_path,
    assert_parquet_file_ids,
):
    """Test execution in the presence of some resume files."""
    ## First, copy over our intermediate files.
    ## This prevents overwriting source-controlled resume files.
    temp_path = os.path.join(tmp_path, "resume", "intermediate")
    shutil.copytree(
        os.path.join(resume_dir, "intermediate"),
        temp_path,
    )
    for file_index in range(0, 5):
        rf.write_mapping_start_key(
            temp_path,
            f"map_{os.path.join(small_sky_parts_dir, f'catalog_0{file_index}_of_05.csv')}",
        )
        rf.write_mapping_done_key(
            temp_path,
            f'map_{os.path.join(small_sky_parts_dir, f"catalog_0{file_index}_of_05.csv")}',
        )
        rf.write_splitting_done_key(
            temp_path,
            f'split_{os.path.join(small_sky_parts_dir, f"catalog_0{file_index}_of_05.csv")}',
        )

    shutil.copytree(
        os.path.join(resume_dir, "Norder=0"),
        os.path.join(tmp_path, "resume", "Norder=0"),
    )

    with pytest.raises(ValueError, match="resume"):
        ## Check that we fail if there are some existing intermediate files
        ImportArguments(
            output_catalog_name="resume",
            input_path=small_sky_parts_dir,
            input_format="csv",
            output_path=tmp_path,
            dask_tmp=tmp_path,
            tmp_dir=tmp_path,
            overwrite=True,
            highest_healpix_order=0,
            pixel_threshold=1000,
            progress_bar=False,
        )

    args = ImportArguments(
        output_catalog_name="resume",
        input_path=small_sky_parts_dir,
        input_format="csv",
        output_path=tmp_path,
        dask_tmp=tmp_path,
        tmp_dir=tmp_path,
        overwrite=True,
        resume=True,
        highest_healpix_order=0,
        pixel_threshold=1000,
        progress_bar=False,
    )

    runner.run(args, dask_client)

    # Check that the catalog metadata file exists
    catalog = Catalog.read_from_hipscat(args.catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == args.catalog_path
    assert catalog.catalog_info.ra_column == "ra"
    assert catalog.catalog_info.dec_column == "dec"
    assert catalog.catalog_info.total_rows == 131
    assert len(catalog.get_pixels()) == 1

    # Check that the catalog parquet file exists and contains correct object IDs
    output_file = os.path.join(
        args.catalog_path, "Norder=0", "Dir=0", "Npix=11.parquet"
    )

    expected_ids = [*range(700, 831)]
    assert_parquet_file_ids(output_file, "id", expected_ids)

    ## Re-running the pipeline with fully done intermediate files
    ## should result in no changes to output files.
    shutil.copytree(
        os.path.join(resume_dir, "intermediate"),
        temp_path,
    )
    rf.set_mapping_done(temp_path)
    rf.set_splitting_done(temp_path)
    rf.set_reducing_done(temp_path)

    args = ImportArguments(
        output_catalog_name="resume",
        input_path=small_sky_parts_dir,
        input_format="csv",
        output_path=tmp_path,
        dask_tmp=tmp_path,
        tmp_dir=tmp_path,
        overwrite=True,
        resume=True,
        highest_healpix_order=0,
        pixel_threshold=1000,
        progress_bar=False,
    )

    runner.run(args, dask_client)

    catalog = Catalog.read_from_hipscat(args.catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == args.catalog_path
    assert catalog.catalog_info.ra_column == "ra"
    assert catalog.catalog_info.dec_column == "dec"
    assert catalog.catalog_info.total_rows == 131
    assert len(catalog.get_pixels()) == 1
    assert_parquet_file_ids(output_file, "id", expected_ids)


@pytest.mark.dask
def test_dask_runner(
    dask_client,
    small_sky_parts_dir,
    assert_parquet_file_ids,
    tmp_path,
):
    """Test basic execution."""
    args = ImportArguments(
        output_catalog_name="small_sky_object_catalog",
        input_path=small_sky_parts_dir,
        input_format="csv",
        output_path=tmp_path,
        dask_tmp=tmp_path,
        highest_healpix_order=1,
        progress_bar=False,
    )

    runner.run(args, dask_client)

    # Check that the catalog metadata file exists
    catalog = Catalog.read_from_hipscat(args.catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == args.catalog_path
    assert catalog.catalog_info.ra_column == "ra"
    assert catalog.catalog_info.dec_column == "dec"
    assert catalog.catalog_info.total_rows == 131
    assert len(catalog.get_pixels()) == 1

    # Check that the catalog parquet file exists and contains correct object IDs
    output_file = os.path.join(
        args.catalog_path, "Norder=0", "Dir=0", "Npix=11.parquet"
    )

    expected_ids = [*range(700, 831)]
    assert_parquet_file_ids(output_file, "id", expected_ids)


@pytest.mark.dask
def test_dask_runner_stats_only(dask_client, small_sky_parts_dir, tmp_path):
    """Test basic execution, without generating catalog parquet outputs."""
    args = ImportArguments(
        output_catalog_name="small_sky",
        input_path=small_sky_parts_dir,
        input_format="csv",
        output_path=tmp_path,
        dask_tmp=tmp_path,
        highest_healpix_order=1,
        progress_bar=False,
        debug_stats_only=True,
    )

    runner.run(args, dask_client)

    metadata_filename = os.path.join(args.catalog_path, "catalog_info.json")
    assert os.path.exists(metadata_filename)

    # Check that the catalog parquet file DOES NOT exist
    output_file = os.path.join(
        args.catalog_path, "Norder=0", "Dir=0", "Npix=11.parquet"
    )

    assert not os.path.exists(output_file)



"""Create columnar index of hipscat table using dask for parallelization"""

import dask.dataframe as dd
import numpy as np
from dask.distributed import progress, wait
from hipscat.io import file_io


def create_index(args):
    """Read primary column, indexing column, and other payload data,
    and write to catalog directory."""
    include_columns = [args.indexing_column]
    if args.extra_columns:
        include_columns.extend(args.extra_columns)
    if args.include_order_pixel:
        include_columns.extend(["Norder", "Dir", "Npix"])

    index_dir = file_io.append_paths_to_pointer(args.catalog_path, "index")

    data = dd.read_parquet(
        path=args.input_catalog_path,
        columns=include_columns,
        engine="pyarrow",
        dataset={"partitioning": "hive"},
    )

    if args.include_order_pixel:
        ## Take out the hive dictionary behavior.
        data["Norder"] = data["Norder"].astype(np.int32)
        data["Dir"] = data["Dir"].astype(np.int32)
        data["Npix"] = data["Npix"].astype(np.int32)
    data = data.reset_index()
    if not args.include_hipscat_index:
        data = data.drop(columns=["_hipscat_index"])
    data = data.repartition(partition_size=args.compute_partition_size)
    data = data.set_index(args.indexing_column)
    result = data.to_parquet(
        path=index_dir,
        engine="pyarrow",
        compute_kwargs={"partition_size": args.compute_partition_size},
    )
    if args.progress_bar:  # pragma: no cover
        progress(result)
    else:
        wait(result)
    return len(data)



"""Test resume file operations"""

import hipscat.pixel_math as hist
import numpy.testing as npt
import pytest

from hipscat_import.catalog.resume_files import (
    clean_resume_files,
    is_mapping_done,
    is_reducing_done,
    read_histogram,
    read_mapping_keys,
    read_reducing_keys,
    set_mapping_done,
    set_reducing_done,
    write_histogram,
    write_mapping_done_key,
    write_mapping_start_key,
    write_reducing_key,
)


def test_mapping_done(tmp_path):
    """Verify expected behavior of mapping done file"""
    assert not is_mapping_done(tmp_path)
    set_mapping_done(tmp_path)
    assert is_mapping_done(tmp_path)

    clean_resume_files(tmp_path)
    assert not is_mapping_done(tmp_path)


def test_reducing_done(tmp_path):
    """Verify expected behavior of reducing done file"""
    assert not is_reducing_done(tmp_path)
    set_reducing_done(tmp_path)
    assert is_reducing_done(tmp_path)

    clean_resume_files(tmp_path)
    assert not is_reducing_done(tmp_path)


def test_read_write_histogram(tmp_path):
    """Test that we can read what we write into a histogram file."""
    empty = hist.empty_histogram(0)
    result = read_histogram(tmp_path, 0)
    npt.assert_array_equal(result, empty)

    expected = hist.empty_histogram(0)
    expected[11] = 131
    write_histogram(tmp_path, expected)
    result = read_histogram(tmp_path, 0)
    npt.assert_array_equal(result, expected)

    clean_resume_files(tmp_path)
    result = read_histogram(tmp_path, 0)
    npt.assert_array_equal(result, empty)


def test_read_write_mapping_keys(tmp_path):
    """Test that we can read what we write into a mapping log file."""
    mapping_keys = read_mapping_keys(tmp_path)
    assert len(mapping_keys) == 0

    write_mapping_start_key(tmp_path, "key1")
    write_mapping_done_key(tmp_path, "key1")
    mapping_keys = read_mapping_keys(tmp_path)
    assert len(mapping_keys) == 1
    assert mapping_keys[0] == "key1"

    write_mapping_start_key(tmp_path, "key2")
    write_mapping_done_key(tmp_path, "key2")
    mapping_keys = read_mapping_keys(tmp_path)
    assert len(mapping_keys) == 2
    assert mapping_keys[0] == "key1"
    assert mapping_keys[1] == "key2"

    clean_resume_files(tmp_path)
    mapping_keys = read_mapping_keys(tmp_path)
    assert len(mapping_keys) == 0


def test_read_write_mapping_keys_corrupt(tmp_path):
    """Test that we can read what we write into a mapping log file."""
    mapping_keys = read_mapping_keys(tmp_path)
    assert len(mapping_keys) == 0

    write_mapping_start_key(tmp_path, "key1")
    write_mapping_done_key(tmp_path, "key1")
    mapping_keys = read_mapping_keys(tmp_path)
    assert len(mapping_keys) == 1
    assert mapping_keys[0] == "key1"

    write_mapping_start_key(tmp_path, "key2")
    with pytest.raises(ValueError, match="logs are corrupted"):
        read_mapping_keys(tmp_path)

    clean_resume_files(tmp_path)
    mapping_keys = read_mapping_keys(tmp_path)
    assert len(mapping_keys) == 0


def test_read_write_reducing_keys(tmp_path):
    """Test that we can read what we write into a reducing log file."""
    reducing_keys = read_reducing_keys(tmp_path)
    assert len(reducing_keys) == 0

    write_reducing_key(tmp_path, "key_1")
    reducing_keys = read_reducing_keys(tmp_path)
    assert len(reducing_keys) == 1
    assert reducing_keys[0] == "key_1"

    write_reducing_key(tmp_path, "key_2")
    reducing_keys = read_reducing_keys(tmp_path)
    assert len(reducing_keys) == 2
    assert reducing_keys[0] == "key_1"
    assert reducing_keys[1] == "key_2"

    clean_resume_files(tmp_path)
    reducing_keys = read_reducing_keys(tmp_path)
    assert len(reducing_keys) == 0



"""Tests of map reduce operations"""
import numpy as np
import numpy.testing as npt
import pandas as pd
import pytest
from hipscat.io import file_io, paths

import hipscat_import.margin_cache.margin_cache as mc
from hipscat_import.margin_cache import MarginCacheArguments

# pylint: disable=protected-access


@pytest.mark.dask(timeout=20)
def test_margin_cache_gen(small_sky_source_catalog, tmp_path, dask_client):
    """Test that margin cache generation works end to end."""
    args = MarginCacheArguments(
        margin_threshold=5.0,
        input_catalog_path=small_sky_source_catalog,
        output_path=tmp_path,
        output_catalog_name="catalog_cache",
    )

    assert args.catalog.catalog_info.ra_column == "source_ra"

    mc.generate_margin_cache(args, dask_client)

    print(args.catalog.partition_info.get_healpix_pixels())

    norder = 1
    npix = 47

    test_file = paths.pixel_catalog_file(
        args.catalog_path, norder, npix
    )

    data = pd.read_parquet(test_file)

    assert len(data) == 4

def test_partition_margin_pixel_pairs(small_sky_source_catalog, tmp_path):
    args = MarginCacheArguments(
        margin_threshold=5.0,
        input_catalog_path=small_sky_source_catalog,
        output_path=tmp_path,
        output_catalog_name="catalog_cache",
    )

    margin_pairs = mc._find_partition_margin_pixel_pairs(
        args.catalog.partition_info.get_healpix_pixels(), args.margin_order
    )

    expected = np.array([725, 733, 757, 765, 727, 735, 759, 767, 469, 192])

    npt.assert_array_equal(margin_pairs.iloc[:10]["margin_pixel"], expected)
    assert len(margin_pairs) == 196


def test_create_margin_directory(small_sky_source_catalog, tmp_path):
    args = MarginCacheArguments(
        margin_threshold=5.0,
        input_catalog_path=small_sky_source_catalog,
        output_path=tmp_path,
        output_catalog_name="catalog_cache",
    )

    mc._create_margin_directory(
        stats=args.catalog.partition_info.get_healpix_pixels(),
        output_path=args.catalog_path,
    )

    output = file_io.append_paths_to_pointer(args.catalog_path, "Norder=0", "Dir=0")
    assert file_io.does_file_or_directory_exist(output)



"""Tests of map reduce operations"""

import os

import hipscat.pixel_math as hist
import numpy as np
import numpy.testing as npt
import pandas as pd
import pyarrow as pa
import pytest

import hipscat_import.catalog.map_reduce as mr
from hipscat_import.catalog.file_readers import get_file_reader


def test_read_empty_filename():
    """Empty file name"""
    with pytest.raises(FileNotFoundError):
        mr.map_to_pixels(
            input_file="",
            file_reader=get_file_reader("parquet"),
            highest_order=10,
            ra_column="ra",
            dec_column="dec",
        )


def test_read_wrong_fileformat(small_sky_file0):
    """CSV file attempting to be read as parquet"""
    with pytest.raises(pa.lib.ArrowInvalid):
        mr.map_to_pixels(
            input_file=small_sky_file0,
            file_reader=get_file_reader("parquet"),
            highest_order=0,
            ra_column="ra_mean",
            dec_column="dec_mean",
        )


def test_read_directory(test_data_dir):
    """Provide directory, not file"""
    with pytest.raises(FileNotFoundError):
        mr.map_to_pixels(
            input_file=test_data_dir,
            file_reader=get_file_reader("parquet"),
            highest_order=0,
            ra_column="ra",
            dec_column="dec",
        )


def test_read_bad_fileformat(blank_data_file):
    """Unsupported file format"""
    with pytest.raises(NotImplementedError):
        mr.map_to_pixels(
            input_file=blank_data_file,
            file_reader=None,
            highest_order=0,
            ra_column="ra",
            dec_column="dec",
        )


def test_read_single_fits(formats_fits):
    """Success case - fits file that exists being read as fits"""
    result = mr.map_to_pixels(
        input_file=formats_fits,
        file_reader=get_file_reader("fits"),
        highest_order=0,
        ra_column="ra",
        dec_column="dec",
    )
    assert len(result) == 12
    expected = hist.empty_histogram(0)
    expected[11] = 131
    npt.assert_array_equal(result, expected)


def test_map_headers_wrong(formats_headers_csv):
    """Test loading the a file with non-default headers (without specifying right headers)"""
    with pytest.raises(ValueError, match="header"):
        mr.map_to_pixels(
            input_file=formats_headers_csv,
            file_reader=get_file_reader("csv"),
            highest_order=0,
            ra_column="ra",
            dec_column="dec",
        )


def test_map_headers(formats_headers_csv):
    """Test loading the a file with non-default headers"""
    result = mr.map_to_pixels(
        input_file=formats_headers_csv,
        file_reader=get_file_reader("csv"),
        highest_order=0,
        ra_column="ra_mean",
        dec_column="dec_mean",
    )

    assert len(result) == 12

    expected = hist.empty_histogram(0)
    expected[11] = 8
    npt.assert_array_equal(result, expected)
    assert (result == expected).all()


def test_split_pixels_headers(formats_headers_csv, assert_parquet_file_ids, tmp_path):
    """Test loading the a file with non-default headers"""
    alignment = np.full(12, None)
    alignment[11] = (0, 11, 131)
    mr.split_pixels(
        input_file=formats_headers_csv,
        file_reader=get_file_reader("csv"),
        highest_order=0,
        ra_column="ra_mean",
        dec_column="dec_mean",
        shard_suffix=0,
        cache_path=tmp_path,
        alignment=alignment,
    )

    file_name = os.path.join(
        tmp_path, "order_0", "dir_0", "pixel_11", "shard_0_0.parquet"
    )
    expected_ids = [*range(700, 708)]
    assert_parquet_file_ids(file_name, "object_id", expected_ids)

    file_name = os.path.join(
        tmp_path, "order_0", "dir_0", "pixel_1", "shard_0_0.parquet"
    )
    assert not os.path.exists(file_name)


def test_map_small_sky_order0(small_sky_single_file):
    """Test loading the small sky catalog and partitioning each object into the same large bucket"""
    result = mr.map_to_pixels(
        input_file=small_sky_single_file,
        file_reader=get_file_reader("csv"),
        highest_order=0,
        ra_column="ra",
        dec_column="dec",
    )

    assert len(result) == 12

    expected = hist.empty_histogram(0)
    expected[11] = 131
    npt.assert_array_equal(result, expected)
    assert (result == expected).all()


def test_map_small_sky_part_order1(small_sky_file0):
    """
    Test loading a small portion of the small sky catalog and
    partitioning objects into four smaller buckets
    """
    result = mr.map_to_pixels(
        input_file=small_sky_file0,
        file_reader=get_file_reader("csv"),
        highest_order=1,
        ra_column="ra",
        dec_column="dec",
    )

    assert len(result) == 48

    expected = hist.empty_histogram(1)
    filled_pixels = [5, 7, 11, 2]
    expected[44:] = filled_pixels[:]
    npt.assert_array_equal(result, expected)
    assert (result == expected).all()


def test_reduce_order0(parquet_shards_dir, assert_parquet_file_ids, tmp_path):
    """Test reducing into one large pixel"""
    mr.reduce_pixel_shards(
        cache_path=parquet_shards_dir,
        destination_pixel_order=0,
        destination_pixel_number=11,
        destination_pixel_size=131,
        output_path=tmp_path,
        add_hipscat_index=True,
        ra_column="ra",
        dec_column="dec",
        id_column="id",
        delete_input_files=False,
    )

    output_file = os.path.join(tmp_path, "Norder=0", "Dir=0", "Npix=11.parquet")

    expected_ids = [*range(700, 831)]
    assert_parquet_file_ids(output_file, "id", expected_ids)


def test_reduce_hipscat_index(parquet_shards_dir, assert_parquet_file_ids, tmp_path):
    """Test reducing with or without a _hipscat_index field"""
    mr.reduce_pixel_shards(
        cache_path=parquet_shards_dir,
        destination_pixel_order=0,
        destination_pixel_number=11,
        destination_pixel_size=131,
        output_path=tmp_path,
        ra_column="ra",
        dec_column="dec",
        id_column="id",
        delete_input_files=False,
    )

    output_file = os.path.join(tmp_path, "Norder=0", "Dir=0", "Npix=11.parquet")

    expected_ids = [*range(700, 831)]
    assert_parquet_file_ids(output_file, "id", expected_ids)
    # expected_indexes = [
    #     13598131468743213056,
    #     13560933976658411520,
    #     13561582046530240512,
    #     13696722494273093632,
    #     13588709332114997248,
    #     13552942781667737600,
    #     13601023174257934336,
    #     13557123557418336256,
    #     13591216801265483776,
    #     13565852277582856192,
    #     13553697461939208192,
    #     13563711661973438464,
    #     13590818251897569280,
    #     13560168899495854080,
    #     13557816572940124160,
    #     13596001812279721984,
    #     13564690156971098112,
    #     13557377060258709504,
    # ]
    # assert_parquet_file_index(output_file, expected_indexes)
    data_frame = pd.read_parquet(output_file, engine="pyarrow")
    assert data_frame.index.name == "_hipscat_index"
    npt.assert_array_equal(
        data_frame.columns,
        ["id", "ra", "dec", "ra_error", "dec_error", "Norder", "Dir", "Npix"],
    )

    mr.reduce_pixel_shards(
        cache_path=parquet_shards_dir,
        destination_pixel_order=0,
        destination_pixel_number=11,
        destination_pixel_size=131,
        output_path=tmp_path,
        add_hipscat_index=False,  ## different from above
        ra_column="ra",
        dec_column="dec",
        id_column="id",
        delete_input_files=False,
    )

    assert_parquet_file_ids(output_file, "id", expected_ids)
    data_frame = pd.read_parquet(output_file, engine="pyarrow")
    ## No index name.
    assert data_frame.index.name is None
    ## Data fields are the same.
    npt.assert_array_equal(
        data_frame.columns,
        ["id", "ra", "dec", "ra_error", "dec_error", "Norder", "Dir", "Npix"],
    )


def test_reduce_bad_expectation(parquet_shards_dir, tmp_path):
    """Test reducing into one large pixel"""
    with pytest.raises(ValueError, match="Unexpected number of objects"):
        mr.reduce_pixel_shards(
            cache_path=parquet_shards_dir,
            destination_pixel_order=0,
            destination_pixel_number=11,
            destination_pixel_size=11,  ## should be 131
            output_path=tmp_path,
            ra_column="ra",
            dec_column="dec",
            id_column="id",
            delete_input_files=False,
        )



"""Test behavior of map reduce methods in association."""

import json
import os

import pandas as pd
import pytest

import hipscat_import.catalog.run_import as runner
from hipscat_import.association.arguments import AssociationArguments
from hipscat_import.association.map_reduce import map_association, reduce_association
from hipscat_import.catalog.arguments import ImportArguments


@pytest.mark.dask(timeout=10)
def test_map_association(
    dask_client, tmp_path, formats_headers_csv, small_sky_object_catalog
):
    """Test association with partially-overlapping dataset.

    This has the added benefit of testing a freshly-minted catalog as input."""
    args = ImportArguments(
        output_catalog_name="subset_catalog",
        input_file_list=[formats_headers_csv],
        input_format="csv",
        output_path=tmp_path,
        ra_column="ra_mean",
        dec_column="dec_mean",
        id_column="object_id",
        dask_tmp=tmp_path,
        highest_healpix_order=1,
        progress_bar=False,
    )
    subset_catalog_path = args.catalog_path

    runner.run(args, dask_client)

    with open(
        os.path.join(subset_catalog_path, "catalog_info.json"), "r", encoding="utf-8"
    ) as metadata_info:
        metadata_keywords = json.load(metadata_info)
        assert metadata_keywords["total_rows"] == 8

    args = AssociationArguments(
        primary_input_catalog_path=small_sky_object_catalog,
        primary_id_column="id",
        primary_join_column="id",
        join_input_catalog_path=subset_catalog_path,
        output_catalog_name="association_inner",
        join_foreign_key="object_id",
        join_id_column="object_id",
        output_path=tmp_path,
        progress_bar=False,
    )

    map_association(args)
    intermediate_partitions_file = os.path.join(
        args.catalog_path, "intermediate", "partitions.csv"
    )
    data_frame = pd.read_csv(intermediate_partitions_file)
    assert data_frame["primary_hipscat_index"].sum() == 8


def test_reduce_bad_inputs(tmp_path, assert_text_file_matches):
    """Test reducing with corrupted input."""

    input_path = os.path.join(tmp_path, "incomplete_inputs")
    os.makedirs(input_path, exist_ok=True)

    output_path = os.path.join(tmp_path, "output")
    os.makedirs(output_path, exist_ok=True)

    ## We don't even have a partitions file
    with pytest.raises(FileNotFoundError):
        reduce_association(input_path=input_path, output_path=output_path)

    ## Create a partitions file, but it doesn't have the right columns
    partitions_data = pd.DataFrame(
        data=[[700, 282.5, -58.5], [701, 299.5, -48.5]],
        columns=["id", "ra", "dec"],
    )
    partitions_csv_file = os.path.join(input_path, "partitions.csv")
    partitions_data.to_csv(partitions_csv_file, index=False)

    with pytest.raises(KeyError, match="primary_hipscat_index"):
        reduce_association(input_path=input_path, output_path=output_path)

    ## Create a partitions file, but it doesn't have corresponding parquet data.
    partitions_data = pd.DataFrame(
        data=[[0, 0, 11, 0, 0, 11, 131]],
        columns=[
            "Norder",
            "Dir",
            "Npix",
            "join_Norder",
            "join_Dir",
            "join_Npix",
            "primary_hipscat_index",
        ],
    )
    partitions_data.to_csv(partitions_csv_file, index=False)

    with pytest.raises(FileNotFoundError):
        reduce_association(input_path=input_path, output_path=output_path)

    ## We still wrote out the partition info file, though!
    expected_lines = [
        "Norder,Dir,Npix,join_Norder,join_Dir,join_Npix,num_rows",
        "0,0,11,0,0,11,131",
    ]
    metadata_filename = os.path.join(output_path, "partition_join_info.csv")
    assert_text_file_matches(expected_lines, metadata_filename)


def test_reduce_bad_expectation(tmp_path):
    """Test reducing with corrupted input."""
    input_path = os.path.join(tmp_path, "incomplete_inputs")
    os.makedirs(input_path, exist_ok=True)

    output_path = os.path.join(tmp_path, "output")
    os.makedirs(output_path, exist_ok=True)

    ## Create a partitions file, and a parquet file with not-enough rows.
    partitions_data = pd.DataFrame(
        data=[[0, 0, 11, 0, 0, 11, 3]],
        columns=[
            "Norder",
            "Dir",
            "Npix",
            "join_Norder",
            "join_Dir",
            "join_Npix",
            "primary_hipscat_index",
        ],
    )
    partitions_csv_file = os.path.join(input_path, "partitions.csv")
    partitions_data.to_csv(partitions_csv_file, index=False)

    parquet_dir = os.path.join(
        input_path,
        "Norder=0",
        "Dir=0",
        "Npix=11",
        "join_Norder=0",
        "join_Dir=0",
        "join_Npix=11",
    )
    os.makedirs(parquet_dir, exist_ok=True)

    parquet_data = pd.DataFrame(
        data=[[700, 7_000_000, 800, 8_000_000], [701, 7_000_100, 801, 8_001_000]],
        columns=[
            "primary_id",
            "primary_hipscat_index",
            "join_id",
            "join_hipscat_index",
        ],
    )
    parquet_data.to_parquet(os.path.join(parquet_dir, "part0.parquet"))
    with pytest.raises(ValueError, match="Unexpected"):
        reduce_association(input_path=input_path, output_path=output_path)

    ## Add one more row in another file, and the expectation is met.
    parquet_data = pd.DataFrame(
        data=[[702, 7_002_000, 802, 8_002_000]],
        columns=[
            "primary_id",
            "primary_hipscat_index",
            "join_id",
            "join_hipscat_index",
        ],
    )
    parquet_data.to_parquet(os.path.join(parquet_dir, "part1.parquet"))
    reduce_association(input_path=input_path, output_path=output_path)



"""All modules for importing new catalogs."""

from .arguments import ImportArguments
from .file_readers import (
    CsvReader,
    FitsReader,
    InputReader,
    ParquetReader,
    get_file_reader,
)
from .map_reduce import map_to_pixels, reduce_pixel_shards, split_pixels
from .resume_files import (
    clean_resume_files,
    is_mapping_done,
    is_reducing_done,
    read_histogram,
    read_mapping_keys,
    read_reducing_keys,
    set_mapping_done,
    set_reducing_done,
    write_histogram,
    write_mapping_done_key,
    write_mapping_start_key,
    write_reducing_key,
)
from .run_import import run



"""Tests of argument validation"""


import pytest

from hipscat_import.association.arguments import AssociationArguments

# pylint: disable=protected-access


def test_none():
    """No arguments provided. Should error for required args."""
    with pytest.raises(ValueError):
        AssociationArguments()


def test_empty_required(tmp_path, small_sky_object_catalog):
    """All non-runtime arguments are required."""
    ## primary_input_catalog_path is missing
    with pytest.raises(ValueError, match="primary_input_catalog_path"):
        AssociationArguments(
            primary_input_catalog_path=None,  ## empty
            primary_id_column="id",
            primary_join_column="id",
            join_input_catalog_path=small_sky_object_catalog,
            join_id_column="id",
            join_foreign_key="id",
            output_path=tmp_path,
            output_catalog_name="small_sky_self_join",
        )

    with pytest.raises(ValueError, match="primary_id_column"):
        AssociationArguments(
            primary_input_catalog_path=small_sky_object_catalog,
            primary_id_column="",  ## empty
            primary_join_column="id",
            join_input_catalog_path=small_sky_object_catalog,
            join_id_column="id",
            join_foreign_key="id",
            output_path=tmp_path,
            output_catalog_name="small_sky_self_join",
            overwrite=True,
        )

    with pytest.raises(ValueError, match="primary_join_column"):
        AssociationArguments(
            primary_input_catalog_path=small_sky_object_catalog,
            primary_id_column="id",
            primary_join_column="",  ## empty
            join_input_catalog_path=small_sky_object_catalog,
            join_id_column="id",
            join_foreign_key="id",
            output_path=tmp_path,
            output_catalog_name="small_sky_self_join",
            overwrite=True,
        )

    with pytest.raises(ValueError, match="join_input_catalog_path"):
        AssociationArguments(
            primary_input_catalog_path=small_sky_object_catalog,
            primary_id_column="id",
            primary_join_column="id",
            join_input_catalog_path="",  ## empty
            join_id_column="id",
            join_foreign_key="id",
            output_path=tmp_path,
            output_catalog_name="small_sky_self_join",
            overwrite=True,
        )

    with pytest.raises(ValueError, match="join_id_column"):
        AssociationArguments(
            primary_input_catalog_path=small_sky_object_catalog,
            primary_id_column="id",
            primary_join_column="id",
            join_input_catalog_path=small_sky_object_catalog,
            join_id_column="",  ## empty
            join_foreign_key="id",
            output_path=tmp_path,
            output_catalog_name="small_sky_self_join",
            overwrite=True,
        )

    with pytest.raises(ValueError, match="join_foreign_key"):
        AssociationArguments(
            primary_input_catalog_path=small_sky_object_catalog,
            primary_id_column="id",
            primary_join_column="id",
            join_input_catalog_path=small_sky_object_catalog,
            join_id_column="id",
            join_foreign_key="",  ## empty
            output_path=tmp_path,
            output_catalog_name="small_sky_self_join",
            overwrite=True,
        )

    with pytest.raises(ValueError, match="output_path"):
        AssociationArguments(
            primary_input_catalog_path=small_sky_object_catalog,
            primary_id_column="id",
            primary_join_column="id",
            join_input_catalog_path=small_sky_object_catalog,
            join_id_column="id",
            join_foreign_key="id",
            output_path="",  ## empty
            output_catalog_name="small_sky_self_join",
            overwrite=True,
        )

    with pytest.raises(ValueError, match="output_catalog_name"):
        AssociationArguments(
            primary_input_catalog_path=small_sky_object_catalog,
            primary_id_column="id",
            primary_join_column="id",
            join_input_catalog_path=small_sky_object_catalog,
            join_id_column="id",
            join_foreign_key="id",
            output_path=tmp_path,
            output_catalog_name="",  ## empty
            overwrite=True,
        )


def test_column_names(tmp_path, small_sky_object_catalog):
    """Test validation of column names."""
    with pytest.raises(ValueError, match="primary_id_column"):
        AssociationArguments(
            primary_input_catalog_path=small_sky_object_catalog,
            primary_id_column="primary_id",
            primary_join_column="id",
            join_input_catalog_path=small_sky_object_catalog,
            join_id_column="id",
            join_foreign_key="id",
            output_path=tmp_path,
            output_catalog_name="bad_columns",  ## empty
            overwrite=True,
        )

    with pytest.raises(ValueError, match="join_id_column"):
        AssociationArguments(
            primary_input_catalog_path=small_sky_object_catalog,
            primary_id_column="id",
            primary_join_column="id",
            join_input_catalog_path=small_sky_object_catalog,
            join_id_column="primary_id",
            join_foreign_key="id",
            output_path=tmp_path,
            output_catalog_name="bad_columns",  ## empty
            overwrite=True,
        )


def test_compute_partition_size(tmp_path, small_sky_object_catalog):
    """Test validation of compute_partition_size."""
    with pytest.raises(ValueError, match="compute_partition_size"):
        AssociationArguments(
            primary_input_catalog_path=small_sky_object_catalog,
            primary_id_column="id",
            primary_join_column="id",
            join_input_catalog_path=small_sky_object_catalog,
            join_id_column="id",
            join_foreign_key="id",
            output_path=tmp_path,
            output_catalog_name="bad_columns",
            compute_partition_size=10,  ## not a valid join option
            overwrite=True,
        )


def test_all_required_args(tmp_path, small_sky_object_catalog):
    """Required arguments are provided."""
    AssociationArguments(
        primary_input_catalog_path=small_sky_object_catalog,
        primary_id_column="id",
        primary_join_column="id",
        join_input_catalog_path=small_sky_object_catalog,
        join_id_column="id",
        join_foreign_key="id",
        output_path=tmp_path,
        output_catalog_name="small_sky_self_join",
    )


def test_to_catalog_info(small_sky_object_catalog, tmp_path):
    """Verify creation of catalog parameters for association table to be created."""
    args = AssociationArguments(
        primary_input_catalog_path=small_sky_object_catalog,
        primary_id_column="id",
        primary_join_column="id",
        join_input_catalog_path=small_sky_object_catalog,
        join_id_column="id",
        join_foreign_key="id",
        output_path=tmp_path,
        output_catalog_name="small_sky_self_join",
    )
    catalog_info = args.to_catalog_info(total_rows=10)
    assert catalog_info.catalog_name == args.output_catalog_name
    assert catalog_info.total_rows == 10


def test_provenance_info(small_sky_object_catalog, tmp_path):
    """Verify that provenance info includes association-specific fields."""
    args = AssociationArguments(
        primary_input_catalog_path=small_sky_object_catalog,
        primary_id_column="id",
        primary_join_column="id",
        join_input_catalog_path=small_sky_object_catalog,
        join_id_column="id",
        join_foreign_key="id",
        output_path=tmp_path,
        output_catalog_name="small_sky_self_join",
    )

    runtime_args = args.provenance_info()["runtime_args"]
    assert "primary_input_catalog_path" in runtime_args



"""test stuff."""

import os

import pyarrow as pa
import pyarrow.parquet as pq
import pytest
from hipscat.catalog.dataset.dataset import Dataset

import hipscat_import.index.run_index as runner
from hipscat_import.index.arguments import IndexArguments


def test_empty_args():
    """Runner should fail with empty arguments"""
    with pytest.raises(TypeError, match="IndexArguments"):
        runner.run(None)


def test_bad_args():
    """Runner should fail with mis-typed arguments"""
    args = {"output_catalog_name": "bad_arg_type"}
    with pytest.raises(TypeError, match="IndexArguments"):
        runner.run(args)


@pytest.mark.dask
def test_run_index(
    small_sky_object_catalog,
    tmp_path,
):
    """Test appropriate metadata is written"""

    args = IndexArguments(
        input_catalog_path=small_sky_object_catalog,
        indexing_column="id",
        output_path=tmp_path,
        output_catalog_name="small_sky_object_index",
        overwrite=True,
        progress_bar=False,
    )
    runner.run(args)

    # Check that the catalog metadata file exists
    catalog = Dataset.read_from_hipscat(args.catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == args.catalog_path

    basic_index_parquet_schema = pa.schema(
        [
            pa.field("_hipscat_index", pa.uint64()),
            pa.field("Norder", pa.int32()),
            pa.field("Dir", pa.int32()),
            pa.field("Npix", pa.int32()),
            pa.field("id", pa.int64()),
        ]
    )
    schema = pq.read_metadata(
        os.path.join(args.catalog_path, "_metadata")
    ).schema.to_arrow_schema()
    assert schema.equals(basic_index_parquet_schema, check_metadata=False)

    schema = pq.read_metadata(
        os.path.join(args.catalog_path, "_common_metadata")
    ).schema.to_arrow_schema()
    assert schema.equals(basic_index_parquet_schema, check_metadata=False)



"""Import a set of non-hipscat files using dask for parallelization

Methods in this file set up a dask pipeline using futures. 
The actual logic of the map reduce is in the `map_reduce.py` file.
"""


import hipscat.io.write_metadata as io
import numpy as np
from dask.distributed import as_completed
from hipscat import pixel_math
from tqdm import tqdm

import hipscat_import.catalog.map_reduce as mr
import hipscat_import.catalog.resume_files as resume
from hipscat_import.catalog.arguments import ImportArguments


def _map_pixels(args, client):
    """Generate a raw histogram of object counts in each healpix pixel"""

    raw_histogram = resume.read_histogram(args.tmp_path, args.mapping_healpix_order)
    if resume.is_mapping_done(args.tmp_path):
        return raw_histogram

    mapped_paths = set(resume.read_mapping_keys(args.tmp_path))

    futures = []
    for file_path in args.input_paths:
        map_key = f"map_{file_path}"
        if map_key in mapped_paths:
            continue
        futures.append(
            client.submit(
                mr.map_to_pixels,
                key=map_key,
                input_file=file_path,
                file_reader=args.file_reader,
                filter_function=args.filter_function,
                highest_order=args.mapping_healpix_order,
                ra_column=args.ra_column,
                dec_column=args.dec_column,
            )
        )

    some_error = False
    for future, result in tqdm(
        as_completed(futures, with_results=True),
        desc="Mapping  ",
        total=len(futures),
        disable=(not args.progress_bar),
    ):
        if future.status == "error":  # pragma: no cover
            some_error = True
            continue
        raw_histogram = np.add(raw_histogram, result)
        resume.write_mapping_start_key(args.tmp_path, future.key)
        resume.write_histogram(args.tmp_path, raw_histogram)
        resume.write_mapping_done_key(args.tmp_path, future.key)
    if some_error:  # pragma: no cover
        raise RuntimeError("Some mapping stages failed. See logs for details.")
    resume.set_mapping_done(args.tmp_path)
    return raw_histogram


def _split_pixels(args, alignment_future, client):
    """Generate a raw histogram of object counts in each healpix pixel"""

    if resume.is_splitting_done(args.tmp_path):
        return

    split_paths = set(resume.read_splitting_keys(args.tmp_path))

    futures = []
    for i, file_path in enumerate(args.input_paths):
        split_key = f"split_{file_path}"
        if split_key in split_paths:
            continue
        futures.append(
            client.submit(
                mr.split_pixels,
                key=split_key,
                input_file=file_path,
                file_reader=args.file_reader,
                filter_function=args.filter_function,
                highest_order=args.mapping_healpix_order,
                ra_column=args.ra_column,
                dec_column=args.dec_column,
                shard_suffix=i,
                cache_path=args.tmp_path,
                alignment=alignment_future,
            )
        )

    some_error = False
    for future in tqdm(
        as_completed(futures),
        desc="Splitting",
        total=len(futures),
        disable=(not args.progress_bar),
    ):
        if future.status == "error":  # pragma: no cover
            some_error = True
            continue
        resume.write_splitting_done_key(args.tmp_path, future.key)
    if some_error:  # pragma: no cover
        raise RuntimeError("Some splitting stages failed. See logs for details.")
    resume.set_splitting_done(args.tmp_path)


def _reduce_pixels(args, destination_pixel_map, client):
    """Loop over destination pixels and merge into parquet files"""

    if resume.is_reducing_done(args.tmp_path):
        return

    reduced_keys = set(resume.read_reducing_keys(args.tmp_path))

    futures = []
    for destination_pixel, source_pixels in destination_pixel_map.items():
        destination_pixel_key = f"{destination_pixel.order}_{destination_pixel.pixel}"
        if destination_pixel_key in reduced_keys:
            continue
        futures.append(
            client.submit(
                mr.reduce_pixel_shards,
                key=destination_pixel_key,
                cache_path=args.tmp_path,
                destination_pixel_order=destination_pixel.order,
                destination_pixel_number=destination_pixel.pixel,
                destination_pixel_size=source_pixels[0],
                output_path=args.catalog_path,
                ra_column=args.ra_column,
                dec_column=args.dec_column,
                id_column=args.id_column,
                add_hipscat_index=args.add_hipscat_index,
                use_schema_file=args.use_schema_file,
            )
        )

    some_error = False
    for future in tqdm(
        as_completed(futures),
        desc="Reducing ",
        total=len(futures),
        disable=(not args.progress_bar),
    ):
        if future.status == "error":  # pragma: no cover
            some_error = True
            continue
        resume.write_reducing_key(args.tmp_path, future.key)
    if some_error:  # pragma: no cover
        raise RuntimeError("Some reducing stages failed. See logs for details.")
    resume.set_reducing_done(args.tmp_path)


def run(args, client):
    """Run catalog creation pipeline."""
    if not args:
        raise ValueError("args is required and should be type ImportArguments")
    if not isinstance(args, ImportArguments):
        raise ValueError("args must be type ImportArguments")
    raw_histogram = _map_pixels(args, client)

    with tqdm(
        total=1, desc="Binning  ", disable=not args.progress_bar
    ) as step_progress:
        if args.constant_healpix_order >= 0:
            alignment = np.full(len(raw_histogram), None)
            for pixel_num, pixel_sum in enumerate(raw_histogram):
                alignment[pixel_num] = (
                    args.constant_healpix_order,
                    pixel_num,
                    pixel_sum,
                )

            destination_pixel_map = pixel_math.generate_constant_pixel_map(
                histogram=raw_histogram,
                constant_healpix_order=args.constant_healpix_order,
            )
        else:
            alignment = pixel_math.generate_alignment(
                raw_histogram,
                highest_order=args.highest_healpix_order,
                threshold=args.pixel_threshold,
            )
            destination_pixel_map = pixel_math.compute_pixel_map(
                raw_histogram,
                highest_order=args.highest_healpix_order,
                threshold=args.pixel_threshold,
            )
        step_progress.update(1)

    if not args.debug_stats_only:
        alignment_future = client.scatter(alignment)
        _split_pixels(args, alignment_future, client)
        _reduce_pixels(args, destination_pixel_map, client)

    # All done - write out the metadata
    with tqdm(
        total=6, desc="Finishing", disable=not args.progress_bar
    ) as step_progress:
        catalog_info = args.to_catalog_info(int(raw_histogram.sum()))
        io.write_provenance_info(
            catalog_base_dir=args.catalog_path,
            dataset_info=catalog_info,
            tool_args=args.provenance_info(),
        )
        step_progress.update(1)

        io.write_catalog_info(
            catalog_base_dir=args.catalog_path, dataset_info=catalog_info
        )
        step_progress.update(1)
        if not args.debug_stats_only:
            io.write_parquet_metadata(args.catalog_path)
        step_progress.update(1)
        io.write_fits_map(args.catalog_path, raw_histogram)
        step_progress.update(1)
        io.write_partition_info(
            catalog_base_dir=args.catalog_path,
            destination_healpix_pixel_map=destination_pixel_map,
        )
        step_progress.update(1)
        resume.clean_resume_files(args.tmp_path)
        step_progress.update(1)



import healpy as hp
import numpy as np
import pandas as pd
import pyarrow.dataset as ds
from hipscat import pixel_math
from hipscat.io import file_io, paths

# pylint: disable=too-many-locals,too-many-arguments


def map_pixel_shards(
    partition_file,
    margin_pairs,
    margin_threshold,
    output_path,
    margin_order,
    ra_column,
    dec_column,
):
    """Creates margin cache shards from a source partition file."""
    data = pd.read_parquet(partition_file)

    data["margin_pixel"] = hp.ang2pix(
        2**margin_order,
        data[ra_column].values,
        data[dec_column].values,
        lonlat=True,
        nest=True,
    )

    constrained_data = data.merge(margin_pairs, on="margin_pixel")

    if len(constrained_data):
        constrained_data.groupby(["partition_order", "partition_pixel"]).apply(
            _to_pixel_shard,
            margin_threshold=margin_threshold,
            output_path=output_path,
            margin_order=margin_order,
            ra_column=ra_column,
            dec_column=dec_column,
        )


def _to_pixel_shard(
    data, margin_threshold, output_path, margin_order, ra_column, dec_column
):
    """Do boundary checking for the cached partition and then output remaining data."""
    order, pix = data["partition_order"].iloc[0], data["partition_pixel"].iloc[0]
    source_order, source_pix = data["Norder"].iloc[0], data["Npix"].iloc[0]

    scale = pixel_math.get_margin_scale(order, margin_threshold)
    bounding_polygons = pixel_math.get_margin_bounds_and_wcs(order, pix, scale)
    is_polar, _ = pixel_math.pixel_is_polar(order, pix)

    if is_polar:
        data = _margin_filter_polar(
            data,
            order,
            pix,
            margin_order,
            margin_threshold,
            ra_column,
            dec_column,
            bounding_polygons,
        )
    else:
        data["margin_check"] = pixel_math.check_margin_bounds(
            data[ra_column].values, data[dec_column].values, bounding_polygons
        )

    # pylint: disable-next=singleton-comparison
    margin_data = data.loc[data["margin_check"] == True]

    if len(margin_data):
        # TODO: this should be a utility function in `hipscat`
        # that properly handles the hive formatting
        # generate a file name for our margin shard
        partition_dir = _get_partition_directory(output_path, order, pix)
        shard_dir = paths.pixel_directory(
            partition_dir, source_order, source_pix
        )

        file_io.make_directory(shard_dir, exist_ok=True)

        shard_path = paths.pixel_catalog_file(partition_dir, source_order, source_pix)

        final_df = margin_data.drop(
            columns=[
                "partition_order",
                "partition_pixel",
                "margin_check",
                "margin_pixel",
            ]
        )

        if is_polar:
            final_df = final_df.drop(columns=["is_trunc"])

        final_df.to_parquet(shard_path)


def _margin_filter_polar(
    data,
    order,
    pix,
    margin_order,
    margin_threshold,
    ra_column,
    dec_column,
    bounding_polygons,
):
    """Filter out margin data around the poles."""
    trunc_pix = pixel_math.get_truncated_margin_pixels(
        order=order, pix=pix, margin_order=margin_order
    )
    data["is_trunc"] = np.isin(data["margin_pixel"], trunc_pix)

    # pylint: disable=singleton-comparison
    trunc_data = data.loc[data["is_trunc"] == True]
    other_data = data.loc[data["is_trunc"] == False]

    trunc_check = pixel_math.check_polar_margin_bounds(
        trunc_data[ra_column].values,
        trunc_data[dec_column].values,
        order,
        pix,
        margin_order,
        margin_threshold,
    )
    data.loc[data["is_trunc"] == True, "margin_check"] = trunc_check

    no_trunc_check = pixel_math.check_margin_bounds(
        other_data[ra_column].values, other_data[dec_column].values, bounding_polygons
    )
    data.loc[data["is_trunc"] == False, "margin_check"] = no_trunc_check
    # pylint: enable=singleton-comparison

    return data

def _get_partition_directory(path, order, pix):
    """Get the directory where a partition pixel's margin shards live"""
    partition_file = paths.pixel_catalog_file(path, order, pix)

    # removes the '.parquet' and adds a slash
    partition_dir = f"{partition_file[:-8]}/"

    return partition_dir

def reduce_margin_shards(output_path, partition_order, partition_pixel):
    """Reduce all partition pixel directories into a single file"""
    shard_dir = _get_partition_directory(
        output_path,
        partition_order,
        partition_pixel
    )

    if file_io.does_file_or_directory_exist(shard_dir):
        data = ds.dataset(shard_dir, format="parquet")
        full_df = data.to_table().to_pandas()

        if len(full_df):
            margin_cache_file_path = paths.pixel_catalog_file(
                output_path,
                partition_order,
                partition_pixel
            )

            full_df.to_parquet(margin_cache_file_path)
            file_io.remove_directory(shard_dir)



"""Utility to hold all arguments required throughout association pipeline"""

from dataclasses import dataclass

from hipscat.catalog.association_catalog.association_catalog import (
    AssociationCatalogInfo,
)

from hipscat_import.runtime_arguments import RuntimeArguments

# pylint: disable=too-many-instance-attributes


@dataclass
class AssociationArguments(RuntimeArguments):
    """Data class for holding association arguments"""

    ## Input - Primary
    primary_input_catalog_path: str = ""
    primary_id_column: str = ""
    primary_join_column: str = ""

    ## Input - Join
    join_input_catalog_path: str = ""
    join_id_column: str = ""
    join_foreign_key: str = ""

    compute_partition_size: int = 1_000_000_000

    def __post_init__(self):
        self._check_arguments()

    def _check_arguments(self):
        super()._check_arguments()
        if not self.primary_input_catalog_path:
            raise ValueError("primary_input_catalog_path is required")
        if not self.primary_id_column:
            raise ValueError("primary_id_column is required")
        if not self.primary_join_column:
            raise ValueError("primary_join_column is required")
        if self.primary_id_column in ["primary_id", "join_id"]:
            raise ValueError("primary_id_column uses a reserved column name")

        if not self.join_input_catalog_path:
            raise ValueError("join_input_catalog_path is required")
        if not self.join_id_column:
            raise ValueError("join_id_column is required")
        if not self.join_foreign_key:
            raise ValueError("join_foreign_key is required")
        if self.join_id_column in ["primary_id", "join_id"]:
            raise ValueError("join_id_column uses a reserved column name")

        if self.compute_partition_size < 100_000:
            raise ValueError("compute_partition_size must be at least 100_000")

    def to_catalog_info(self, total_rows) -> AssociationCatalogInfo:
        """Catalog-type-specific dataset info."""
        info = {
            "catalog_name": self.output_catalog_name,
            "catalog_type": "association",
            "total_rows": total_rows,
            "primary_column": self.primary_id_column,
            "primary_catalog": str(self.primary_input_catalog_path),
            "join_column": self.join_id_column,
            "join_catalog": str(self.join_input_catalog_path),
        }
        return AssociationCatalogInfo(**info)

    def additional_runtime_provenance_info(self) -> dict:
        return {
            "primary_input_catalog_path": str(self.primary_input_catalog_path),
            "primary_id_column": self.primary_id_column,
            "primary_join_column": self.primary_join_column,
            "join_input_catalog_path": str(self.join_input_catalog_path),
            "join_id_column": self.join_id_column,
            "join_foreign_key": self.join_foreign_key,
        }



"""Methods for reading and writing intermediate files for import execution."""

from datetime import datetime
from pathlib import Path

import numpy as np
import pandas as pd
from hipscat import pixel_math
from hipscat.io import FilePointer, file_io
from numpy import frombuffer


def read_histogram(tmp_path: FilePointer, highest_healpix_order):
    """Read a numpy array at the indicated directory.
    Otherwise, return histogram of appropriate shape."""
    file_name = file_io.append_paths_to_pointer(tmp_path, "mapping_histogram.binary")
    if file_io.does_file_or_directory_exist(file_name):
        with open(file_name, "rb") as file_handle:
            return frombuffer(file_handle.read(), dtype=np.int64)
    return pixel_math.empty_histogram(highest_healpix_order)


def write_histogram(tmp_path: FilePointer, raw_histogram):
    """overwrite existing raw histogram with updated values."""
    file_name = file_io.append_paths_to_pointer(tmp_path, "mapping_histogram.binary")
    with open(file_name, "wb+") as file_handle:
        file_handle.write(raw_histogram.data)


def is_mapping_done(tmp_path: FilePointer):
    """Check for existence of done file"""
    return file_io.does_file_or_directory_exist(
        file_io.append_paths_to_pointer(tmp_path, "mapping_done")
    )


def set_mapping_done(tmp_path: FilePointer):
    """Touch (create) a done file"""
    Path(file_io.append_paths_to_pointer(tmp_path, "mapping_done")).touch()


def is_splitting_done(tmp_path: FilePointer):
    """Check for existence of done file"""
    return file_io.does_file_or_directory_exist(
        file_io.append_paths_to_pointer(tmp_path, "splitting_done")
    )


def set_splitting_done(tmp_path: FilePointer):
    """Touch (create) a done file"""
    Path(file_io.append_paths_to_pointer(tmp_path, "splitting_done")).touch()


def is_reducing_done(tmp_path: FilePointer):
    """Check for existence of done file"""
    return file_io.does_file_or_directory_exist(
        file_io.append_paths_to_pointer(tmp_path, "reducing_done")
    )


def set_reducing_done(tmp_path: FilePointer):
    """Touch (create) a done file"""
    Path(file_io.append_paths_to_pointer(tmp_path, "reducing_done")).touch()


def read_mapping_keys(tmp_path: FilePointer):
    """Read keys from mapping log file"""
    mapping_start_keys = _read_log_keys(
        file_io.append_paths_to_pointer(tmp_path, "mapping_start_log.txt")
    )
    mapping_done_keys = _read_log_keys(
        file_io.append_paths_to_pointer(tmp_path, "mapping_done_log.txt")
    )
    if len(mapping_done_keys) != len(mapping_start_keys):
        raise ValueError(
            "Resume logs are corrupted. Delete temp directory and restart import pipeline."
        )
    return mapping_start_keys


def read_splitting_keys(tmp_path: FilePointer):
    """Read keys from splitting log file"""
    return _read_log_keys(
        file_io.append_paths_to_pointer(tmp_path, "splitting_done_log.txt")
    )


def read_reducing_keys(tmp_path: FilePointer):
    """Read keys from reducing log file"""
    return _read_log_keys(file_io.append_paths_to_pointer(tmp_path, "reducing_log.txt"))


def _read_log_keys(file_name):
    """Read a resume log file, containing timestamp and keys."""
    if file_io.does_file_or_directory_exist(file_name):
        mapping_log = pd.read_csv(
            file_name,
            delimiter="\t",
            header=None,
            names=["time", "key"],
        )
        return mapping_log["key"].tolist()
    return []


def write_mapping_start_key(tmp_path: FilePointer, key):
    """Append single key to mapping log file"""
    _write_log_key(
        file_io.append_paths_to_pointer(tmp_path, "mapping_start_log.txt"), key
    )


def write_mapping_done_key(tmp_path: FilePointer, key):
    """Append single key to mapping log file"""
    _write_log_key(
        file_io.append_paths_to_pointer(tmp_path, "mapping_done_log.txt"), key
    )


def write_splitting_done_key(tmp_path: FilePointer, key):
    """Append single key to splitting log file"""
    _write_log_key(
        file_io.append_paths_to_pointer(tmp_path, "splitting_done_log.txt"), key
    )


def write_reducing_key(tmp_path: FilePointer, key):
    """Append single key to reducing log file"""
    _write_log_key(file_io.append_paths_to_pointer(tmp_path, "reducing_log.txt"), key)


def _write_log_key(file_name, key):
    """Append a tab-delimited line to the file with the current timestamp and provided key"""
    with open(file_name, "a", encoding="utf-8") as mapping_log:
        mapping_log.write(f'{datetime.now().strftime("%m/%d/%Y, %H:%M:%S")}\t{key}\n')


def clean_resume_files(tmp_path: FilePointer):
    """Remove all intermediate files created in execution."""
    file_io.remove_directory(tmp_path, ignore_errors=True)



"""Create partitioned association table between two catalogs"""

import dask.dataframe as dd
import pyarrow.parquet as pq
from hipscat.io import file_io, paths


def map_association(args):
    """Using dask dataframes, create an association between two catalogs.
    This will write out sharded parquet files to the temp (intermediate)
    directory.

    Implementation notes:

    Because we may be joining to a column that is NOT the natural/primary key
    (on either side of the join), we fetch both the identifying column and the
    join predicate column, possibly duplicating one of the columns.

    This way, when we drop the join predicate columns at the end of the process,
    we will still have the identifying columns. However, it makes the loading
    of each input catalog more verbose.
    """
    ## Read and massage primary input data
    ## NB: We may be joining on a column that is NOT the natural primary key.
    single_primary_column = args.primary_id_column == args.primary_join_column
    read_columns = [
        "Norder",
        "Dir",
        "Npix",
    ]
    if single_primary_column:
        read_columns = [args.primary_id_column] + read_columns
    else:
        read_columns = [args.primary_join_column, args.primary_id_column] + read_columns

    primary_index = dd.read_parquet(
        path=args.primary_input_catalog_path,
        columns=read_columns,
        dataset={"partitioning": "hive"},
    )
    if single_primary_column:
        ## Duplicate the column to simplify later steps
        primary_index["primary_id"] = primary_index[args.primary_join_column]
    rename_columns = {
        args.primary_join_column: "primary_join",
        "_hipscat_index": "primary_hipscat_index",
    }
    if not single_primary_column:
        rename_columns[args.primary_id_column] = "primary_id"
    primary_index = (
        primary_index.reset_index()
        .rename(columns=rename_columns)
        .set_index("primary_join")
    )

    ## Read and massage join input data
    single_join_column = args.join_id_column == args.join_foreign_key
    read_columns = [
        "Norder",
        "Dir",
        "Npix",
    ]
    if single_join_column:
        read_columns = [args.join_id_column] + read_columns
    else:
        read_columns = [args.join_id_column, args.join_foreign_key] + read_columns

    join_index = dd.read_parquet(
        path=args.join_input_catalog_path,
        columns=read_columns,
        dataset={"partitioning": "hive"},
    )
    if single_join_column:
        ## Duplicate the column to simplify later steps
        join_index["join_id"] = join_index[args.join_id_column]
    rename_columns = {
        args.join_foreign_key: "join_to_primary",
        "_hipscat_index": "join_hipscat_index",
        "Norder": "join_Norder",
        "Dir": "join_Dir",
        "Npix": "join_Npix",
    }
    if not single_join_column:
        rename_columns[args.join_id_column] = "join_id"
    join_index = (
        join_index.reset_index()
        .rename(columns=rename_columns)
        .set_index("join_to_primary")
    )

    ## Join the two data sets on the shared join predicate.
    join_data = primary_index.merge(
        join_index, how="inner", left_index=True, right_index=True
    )

    ## Write out a summary of each partition join
    groups = (
        join_data.groupby(
            ["Norder", "Dir", "Npix", "join_Norder", "join_Dir", "join_Npix"],
            group_keys=False,
        )["primary_hipscat_index"]
        .count()
        .compute()
    )
    intermediate_partitions_file = file_io.append_paths_to_pointer(
        args.tmp_path, "partitions.csv"
    )
    file_io.write_dataframe_to_csv(
        dataframe=groups, file_pointer=intermediate_partitions_file
    )

    ## Drop join predicate columns
    join_data = join_data[
        [
            "Norder",
            "Dir",
            "Npix",
            "join_Norder",
            "join_Dir",
            "join_Npix",
            "primary_id",
            "primary_hipscat_index",
            "join_id",
            "join_hipscat_index",
        ]
    ]

    ## Write out association table shards.
    join_data.to_parquet(
        path=args.tmp_path,
        engine="pyarrow",
        partition_on=["Norder", "Dir", "Npix", "join_Norder", "join_Dir", "join_Npix"],
        compute_kwargs={"partition_size": args.compute_partition_size},
        write_index=False,
    )


def reduce_association(input_path, output_path):
    """Collate sharded parquet files into a single parquet file per partition"""
    intermediate_partitions_file = file_io.append_paths_to_pointer(
        input_path, "partitions.csv"
    )
    data_frame = file_io.load_csv_to_pandas(intermediate_partitions_file)

    ## Clean up the dataframe and write out as our new partition join info file.
    data_frame = data_frame[data_frame["primary_hipscat_index"] != 0]
    data_frame["num_rows"] = data_frame["primary_hipscat_index"]
    data_frame = data_frame[
        ["Norder", "Dir", "Npix", "join_Norder", "join_Dir", "join_Npix", "num_rows"]
    ]
    data_frame = data_frame.sort_values(["Norder", "Npix", "join_Norder", "join_Npix"])
    file_io.write_dataframe_to_csv(
        dataframe=data_frame,
        file_pointer=file_io.append_paths_to_pointer(
            output_path, "partition_join_info.csv"
        ),
        index=False,
    )

    ## For each partition, join all parquet shards into single parquet file.
    for _, partition in data_frame.iterrows():
        input_dir = paths.create_hive_directory_name(
            input_path,
            ["Norder", "Dir", "Npix", "join_Norder", "join_Dir", "join_Npix"],
            [
                partition["Norder"],
                partition["Dir"],
                partition["Npix"],
                partition["join_Norder"],
                partition["join_Dir"],
                partition["join_Npix"],
            ],
        )
        output_dir = paths.pixel_association_directory(
            output_path,
            partition["Norder"],
            partition["Npix"],
            partition["join_Norder"],
            partition["join_Npix"],
        )
        file_io.make_directory(output_dir, exist_ok=True)
        output_file = paths.pixel_association_file(
            output_path,
            partition["Norder"],
            partition["Npix"],
            partition["join_Norder"],
            partition["join_Npix"],
        )
        table = pq.read_table(input_dir)
        rows_written = len(table)

        if rows_written != partition["num_rows"]:
            raise ValueError(
                "Unexpected number of objects ",
                f" Expected {partition['num_rows']}, wrote {rows_written}",
            )

        table.to_pandas().set_index("primary_hipscat_index").sort_index().to_parquet(
            output_file
        )

    return data_frame["num_rows"].sum()



"""Create partitioned association table between two catalogs
using dask dataframes for parallelization

Methods in this file set up a dask pipeline using dataframes. 
The actual logic of the map reduce is in the `map_reduce.py` file.
"""

from hipscat.io import file_io, write_metadata
from tqdm import tqdm

from hipscat_import.association.arguments import AssociationArguments
from hipscat_import.association.map_reduce import map_association, reduce_association


def run(args):
    """Run the association pipeline"""
    if not args:
        raise TypeError("args is required and should be type AssociationArguments")
    if not isinstance(args, AssociationArguments):
        raise TypeError("args must be type AssociationArguments")

    with tqdm(total=1, desc="Mapping ", disable=not args.progress_bar) as step_progress:
        map_association(args)
        step_progress.update(1)

    rows_written = 0
    with tqdm(
        total=1, desc="Reducing ", disable=not args.progress_bar
    ) as step_progress:
        rows_written = reduce_association(args.tmp_path, args.catalog_path)
        step_progress.update(1)

    # All done - write out the metadata
    with tqdm(
        total=4, desc="Finishing", disable=not args.progress_bar
    ) as step_progress:
        # pylint: disable=duplicate-code
        # Very similar to /index/run_index.py
        catalog_info = args.to_catalog_info(int(rows_written))
        write_metadata.write_provenance_info(
            catalog_base_dir=args.catalog_path,
            dataset_info=catalog_info,
            tool_args=args.provenance_info(),
        )
        step_progress.update(1)
        catalog_info = args.to_catalog_info(total_rows=int(rows_written))
        write_metadata.write_catalog_info(
            dataset_info=catalog_info, catalog_base_dir=args.catalog_path
        )
        step_progress.update(1)
        write_metadata.write_parquet_metadata(args.catalog_path)
        step_progress.update(1)
        file_io.remove_directory(args.tmp_path, ignore_errors=True)
        step_progress.update(1)



"""Import a set of non-hipscat files using dask for parallelization"""

import healpy as hp
import numpy as np
import pyarrow as pa
import pyarrow.parquet as pq
from hipscat import pixel_math
from hipscat.io import FilePointer, file_io, paths

from hipscat_import.catalog.file_readers import InputReader

# pylint: disable=too-many-locals,too-many-arguments


def _get_pixel_directory(cache_path: FilePointer, order: np.int64, pixel: np.int64):
    """Create a path for intermediate pixel data.

    This will take the form:

        <cache_path>/dir_<directory separator>/pixel_<pixel>

    where the directory separator is calculated using integer division:

        (pixel/10000)*10000

    and exists to mitigate problems on file systems that don't support
    more than 10_000 children nodes.
    """
    dir_number = int(pixel / 10_000) * 10_000
    return file_io.append_paths_to_pointer(
        cache_path, f"order_{order}", f"dir_{dir_number}", f"pixel_{pixel}"
    )


def _has_named_index(dataframe):
    """Heuristic to determine if a dataframe has some meaningful index.

    This will reject dataframes with no index name for a single index,
    or empty names for multi-index (e.g. [] or [None]).
    """
    if dataframe.index.name is not None:
        ## Single index with a given name.
        return True
    if len(dataframe.index.names) == 0 or all(
        name is None for name in dataframe.index.names
    ):
        return False
    return True


def _iterate_input_file(
    input_file: FilePointer,
    file_reader: InputReader,
    highest_order,
    ra_column,
    dec_column,
    filter_function=None,
):
    """Helper function to handle input file reading and healpix pixel calculation"""
    if not file_io.does_file_or_directory_exist(input_file):
        raise FileNotFoundError(f"File not found at path: {input_file}")
    if not file_io.is_regular_file(input_file):
        raise FileNotFoundError(
            f"Directory found at path - requires regular file: {input_file}"
        )
    if not file_reader:
        raise NotImplementedError("No file reader implemented")

    required_columns = [ra_column, dec_column]

    for chunk_number, data in enumerate(file_reader.read(input_file)):
        if not all(x in data.columns for x in required_columns):
            raise ValueError(
                f"Invalid column names in input file: {ra_column}, {dec_column} not in {input_file}"
            )
        # Set up the data we want (filter and find pixel)
        if filter_function:
            data = filter_function(data)
        mapped_pixels = hp.ang2pix(
            2**highest_order,
            data[ra_column].values,
            data[dec_column].values,
            lonlat=True,
            nest=True,
        )
        yield chunk_number, data, mapped_pixels


def map_to_pixels(
    input_file: FilePointer,
    file_reader: InputReader,
    highest_order,
    ra_column,
    dec_column,
    filter_function=None,
):
    """Map a file of input objects to their healpix pixels.

    Args:
        input_file (FilePointer): file to read for catalog data.
        file_reader (hipscat_import.catalog.file_readers.InputReader): instance of input
            reader that specifies arguments necessary for reading from the input file.
        shard_suffix (str): unique counter for this input file, used
            when creating intermediate files
        highest_order (int): healpix order to use when mapping
        ra_column (str): where to find right ascension data in the dataframe
        dec_column (str): where to find declation in the dataframe

    Returns:
        one-dimensional numpy array of long integers where the value at each index corresponds
        to the number of objects found at the healpix pixel.
    Raises:
        ValueError: if the `ra_column` or `dec_column` cannot be found in the input file.
        FileNotFoundError: if the file does not exist, or is a directory
    """
    histo = pixel_math.empty_histogram(highest_order)
    for _, _, mapped_pixels in _iterate_input_file(
        input_file, file_reader, highest_order, ra_column, dec_column, filter_function
    ):
        mapped_pixel, count_at_pixel = np.unique(mapped_pixels, return_counts=True)
        histo[mapped_pixel] += count_at_pixel.astype(np.int64)
    return histo


def split_pixels(
    input_file: FilePointer,
    file_reader: InputReader,
    shard_suffix,
    highest_order,
    ra_column,
    dec_column,
    cache_path: FilePointer,
    alignment=None,
    filter_function=None,
):
    """Map a file of input objects to their healpix pixels and split into shards.

    Args:
        input_file (FilePointer): file to read for catalog data.
        file_reader (hipscat_import.catalog.file_readers.InputReader): instance
            of input reader that specifies arguments necessary for reading from the input file.
        shard_suffix (str): unique counter for this input file, used
            when creating intermediate files
        highest_order (int): healpix order to use when mapping
        ra_column (str): where to find right ascension data in the dataframe
        dec_column (str): where to find declation in the dataframe
        cache_path (FilePointer): where to write intermediate files.
        filter_function (function pointer): method to perform some filtering
            or transformation of the input data

    Raises:
        ValueError: if the `ra_column` or `dec_column` cannot be found in the input file.
        FileNotFoundError: if the file does not exist, or is a directory
    """
    for chunk_number, data, mapped_pixels in _iterate_input_file(
        input_file, file_reader, highest_order, ra_column, dec_column, filter_function
    ):
        aligned_pixels = alignment[mapped_pixels]
        unique_pixels, unique_inverse = np.unique(aligned_pixels, return_inverse=True)

        for unique_index, [order, pixel, _] in enumerate(unique_pixels):
            mapped_indexes = np.where(unique_inverse == unique_index)
            data_indexes = data.index[mapped_indexes[0].tolist()]

            filtered_data = data.filter(items=data_indexes, axis=0)

            pixel_dir = _get_pixel_directory(cache_path, order, pixel)
            file_io.make_directory(pixel_dir, exist_ok=True)
            output_file = file_io.append_paths_to_pointer(
                pixel_dir, f"shard_{shard_suffix}_{chunk_number}.parquet"
            )
            if _has_named_index(filtered_data):
                filtered_data.to_parquet(output_file, index=True)
            else:
                filtered_data.to_parquet(output_file, index=False)
        del filtered_data, data_indexes


def reduce_pixel_shards(
    cache_path,
    destination_pixel_order,
    destination_pixel_number,
    destination_pixel_size,
    output_path,
    ra_column,
    dec_column,
    id_column,
    add_hipscat_index=True,
    delete_input_files=True,
    use_schema_file="",
):
    """Reduce sharded source pixels into destination pixels.

    In addition to combining multiple shards of data into a single
    parquet file, this method will add a few new columns:

        - ``Norder`` - the healpix order for the pixel
        - ``Dir`` - the directory part, corresponding to the pixel
        - ``Npix`` - the healpix pixel
        - ``_hipscat_index`` - optional - a spatially-correlated
          64-bit index field.

    Notes on ``_hipscat_index``:

        - if we generate the field, we will promote any previous
          *named* pandas index field(s) to a column with
          that name.
        - see ``hipscat.pixel_math.hipscat_id``
          for more in-depth discussion of this field.

    Args:
        cache_path (str): where to read intermediate files
        origin_pixel_numbers (list[int]): high order pixels, with object
            data written to intermediate directories.
        destination_pixel_order (int): order of the final catalog pixel
        destination_pixel_number (int): pixel number at the above order
        destination_pixel_size (int): expected number of rows to write
            for the catalog's final pixel
        output_path (str): where to write the final catalog pixel data
        id_column (str): column for survey identifier, or other sortable column
        add_hipscat_index (bool): should we add a _hipscat_index column to
            the resulting parquet file?
        delete_input_files (bool): should we delete the intermediate files
            used as input for this method.
        use_schema_file (str): use the parquet schema from the indicated
            parquet file.

    Raises:
        ValueError: if the number of rows written doesn't equal provided
            `destination_pixel_size`
    """
    destination_dir = paths.pixel_directory(
        output_path, destination_pixel_order, destination_pixel_number
    )
    file_io.make_directory(destination_dir, exist_ok=True)

    destination_file = paths.pixel_catalog_file(
        output_path, destination_pixel_order, destination_pixel_number
    )

    schema = None
    if use_schema_file:
        schema = file_io.read_parquet_metadata(use_schema_file).schema.to_arrow_schema()

    tables = []
    pixel_dir = _get_pixel_directory(
        cache_path, destination_pixel_order, destination_pixel_number
    )

    if schema:
        tables.append(pq.read_table(pixel_dir, schema=schema))
    else:
        tables.append(pq.read_table(pixel_dir))

    merged_table = pa.concat_tables(tables)

    rows_written = len(merged_table)

    if rows_written != destination_pixel_size:
        raise ValueError(
            "Unexpected number of objects at pixel "
            f"({destination_pixel_order}, {destination_pixel_number})."
            f" Expected {destination_pixel_size}, wrote {rows_written}"
        )

    dataframe = merged_table.to_pandas()
    if id_column:
        dataframe = dataframe.sort_values(id_column)
    if add_hipscat_index:
        dataframe["_hipscat_index"] = pixel_math.compute_hipscat_id(
            dataframe[ra_column].values,
            dataframe[dec_column].values,
        )

    dataframe["Norder"] = np.full(
        rows_written, fill_value=destination_pixel_order, dtype=np.int32
    )
    dataframe["Dir"] = np.full(
        rows_written,
        fill_value=int(destination_pixel_number / 10_000) * 10_000,
        dtype=np.int32,
    )
    dataframe["Npix"] = np.full(
        rows_written, fill_value=destination_pixel_number, dtype=np.int32
    )

    if add_hipscat_index:
        ## If we had a meaningful index before, preserve it as a column.
        if _has_named_index(dataframe):
            dataframe = dataframe.reset_index()
        dataframe = dataframe.set_index("_hipscat_index").sort_index()
    dataframe.to_parquet(destination_file)

    del dataframe, merged_table, tables

    if delete_input_files:
        pixel_dir = _get_pixel_directory(
            cache_path, destination_pixel_order, destination_pixel_number
        )

        file_io.remove_directory(pixel_dir, ignore_errors=True)



"""test stuff."""

import os

import numpy as np
import numpy.testing as npt
import pandas as pd
import pytest
from hipscat.catalog.association_catalog.association_catalog import AssociationCatalog

import hipscat_import.association.run_association as runner
from hipscat_import.association.arguments import AssociationArguments


def test_empty_args():
    """Runner should fail with empty arguments"""
    with pytest.raises(TypeError):
        runner.run(None)


def test_bad_args():
    """Runner should fail with mis-typed arguments"""
    args = {"output_catalog_name": "bad_arg_type"}
    with pytest.raises(TypeError):
        runner.run(args)


@pytest.mark.dask
def test_object_to_source(
    small_sky_object_catalog,
    small_sky_source_catalog,
    tmp_path,
):
    """Test creating association between object and source catalogs."""

    args = AssociationArguments(
        primary_input_catalog_path=small_sky_object_catalog,
        primary_id_column="id",
        primary_join_column="id",
        join_input_catalog_path=small_sky_source_catalog,
        output_catalog_name="small_sky_association",
        join_id_column="source_id",
        join_foreign_key="object_id",
        output_path=tmp_path,
        progress_bar=False,
    )
    runner.run(args)

    ## Check that the association data can be parsed as a valid association catalog.
    catalog = AssociationCatalog.read_from_hipscat(args.catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == args.catalog_path
    assert len(catalog.get_join_pixels()) == 14

    ## Test one pixel that will have 50 rows in it.
    output_file = os.path.join(
        tmp_path,
        "small_sky_association",
        "Norder=0",
        "Dir=0",
        "Npix=11",
        "join_Norder=0",
        "join_Dir=0",
        "join_Npix=4.parquet",
    )
    assert os.path.exists(output_file), f"file not found [{output_file}]"
    data_frame = pd.read_parquet(output_file, engine="pyarrow")
    npt.assert_array_equal(
        data_frame.columns,
        ["primary_id", "join_id", "join_hipscat_index"],
    )
    assert data_frame.index.name == "primary_hipscat_index"
    assert len(data_frame) == 50
    ids = data_frame["primary_id"]
    assert np.logical_and(ids >= 700, ids < 832).all()
    ids = data_frame["join_id"]
    assert np.logical_and(ids >= 70_000, ids < 87161).all()

    catalog = AssociationCatalog.read_from_hipscat(args.catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == args.catalog_path
    assert len(catalog.get_join_pixels()) == 14


@pytest.mark.dask
def test_source_to_object(
    small_sky_object_catalog,
    small_sky_source_catalog,
    tmp_path,
):
    """Test creating (weirder) association between source and object catalogs."""

    args = AssociationArguments(
        primary_input_catalog_path=small_sky_source_catalog,
        primary_id_column="source_id",
        primary_join_column="object_id",
        join_input_catalog_path=small_sky_object_catalog,
        join_id_column="id",
        join_foreign_key="id",
        output_catalog_name="small_sky_association",
        output_path=tmp_path,
        progress_bar=False,
    )
    runner.run(args)

    ## Check that the association data can be parsed as a valid association catalog.
    catalog = AssociationCatalog.read_from_hipscat(args.catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == args.catalog_path
    assert len(catalog.get_join_pixels()) == 14

    ## Test one pixel that will have 50 rows in it.
    output_file = os.path.join(
        tmp_path,
        "small_sky_association",
        "Norder=0",
        "Dir=0",
        "Npix=4",
        "join_Norder=0",
        "join_Dir=0",
        "join_Npix=11.parquet",
    )
    assert os.path.exists(output_file), f"file not found [{output_file}]"
    data_frame = pd.read_parquet(output_file, engine="pyarrow")
    npt.assert_array_equal(
        data_frame.columns,
        ["primary_id", "join_id", "join_hipscat_index"],
    )
    assert data_frame.index.name == "primary_hipscat_index"
    assert len(data_frame) == 50
    ids = data_frame["primary_id"]
    assert np.logical_and(ids >= 70_000, ids < 87161).all()
    ids = data_frame["join_id"]
    assert np.logical_and(ids >= 700, ids < 832).all()


@pytest.mark.dask
def test_self_join(
    small_sky_object_catalog,
    tmp_path,
):
    """Test creating association between object catalog and itself."""

    args = AssociationArguments(
        primary_input_catalog_path=small_sky_object_catalog,
        primary_id_column="id",
        primary_join_column="id",
        join_input_catalog_path=small_sky_object_catalog,
        output_catalog_name="small_sky_self_association",
        join_foreign_key="id",
        join_id_column="id",
        output_path=tmp_path,
        progress_bar=False,
    )
    runner.run(args)

    ## Check that the association data can be parsed as a valid association catalog.
    catalog = AssociationCatalog.read_from_hipscat(args.catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == args.catalog_path
    assert len(catalog.get_join_pixels()) == 1

    ## Test one pixel that will have 50 rows in it.
    output_file = os.path.join(
        tmp_path,
        "small_sky_self_association",
        "Norder=0",
        "Dir=0",
        "Npix=11",
        "join_Norder=0",
        "join_Dir=0",
        "join_Npix=11.parquet",
    )
    assert os.path.exists(output_file), f"file not found [{output_file}]"
    data_frame = pd.read_parquet(output_file, engine="pyarrow")
    npt.assert_array_equal(
        data_frame.columns,
        ["primary_id", "join_id", "join_hipscat_index"],
    )
    assert data_frame.index.name == "primary_hipscat_index"
    assert len(data_frame) == 131
    ids = data_frame["primary_id"]
    assert np.logical_and(ids >= 700, ids < 832).all()
    ids = data_frame["join_id"]
    assert np.logical_and(ids >= 700, ids < 832).all()



import pandas as pd
import pytest
from hipscat.io import file_io, paths

from hipscat_import.margin_cache import margin_cache_map_reduce

keep_cols = ["weird_ra", "weird_dec"]

drop_cols = [
    "partition_order",
    "partition_pixel",
    "margin_check",
    "margin_pixel",
    "is_trunc",
]


def validate_result_dataframe(df_path, expected_len):
    res_df = pd.read_parquet(df_path)

    assert len(res_df) == expected_len

    cols = res_df.columns.values.tolist()

    for col in keep_cols:
        assert col in cols

    for col in drop_cols:
        assert col not in cols

@pytest.mark.timeout(5)
def test_to_pixel_shard_equator(tmp_path, basic_data_shard_df):
    margin_cache_map_reduce._to_pixel_shard(
        basic_data_shard_df,
        margin_threshold=0.1,
        output_path=tmp_path,
        margin_order=3,
        ra_column="weird_ra",
        dec_column="weird_dec",
    )

    path = file_io.append_paths_to_pointer(
        tmp_path, "Norder=1/Dir=0/Npix=21/Norder=1/Dir=0/Npix=0.parquet"
    )

    assert file_io.does_file_or_directory_exist(path)

    validate_result_dataframe(path, 46)

@pytest.mark.timeout(5)
def test_to_pixel_shard_polar(tmp_path, polar_data_shard_df):
    margin_cache_map_reduce._to_pixel_shard(
        polar_data_shard_df,
        margin_threshold=0.1,
        output_path=tmp_path,
        margin_order=3,
        ra_column="weird_ra",
        dec_column="weird_dec",
    )

    path = file_io.append_paths_to_pointer(
        tmp_path, "Norder=2/Dir=0/Npix=15/Norder=2/Dir=0/Npix=0.parquet"
    )

    assert file_io.does_file_or_directory_exist(path)

    validate_result_dataframe(path, 317)

def test_reduce_margin_shards(tmp_path, basic_data_shard_df):
    partition_dir = margin_cache_map_reduce._get_partition_directory(
        tmp_path, 1, 21
    )
    shard_dir = paths.pixel_directory(
        partition_dir, 1, 21
    )

    file_io.make_directory(shard_dir, exist_ok=True)

    first_shard_path = paths.pixel_catalog_file(
        partition_dir, 1, 0
    )
    second_shard_path = paths.pixel_catalog_file(
        partition_dir, 1, 1
    )

    print(first_shard_path)

    shard_df = basic_data_shard_df.drop(columns=[
        "partition_order", 
        "partition_pixel",
        "margin_pixel"
    ])

    shard_df.to_parquet(first_shard_path)
    shard_df.to_parquet(second_shard_path)

    margin_cache_map_reduce.reduce_margin_shards(tmp_path, 1, 21)

    result_path = paths.pixel_catalog_file(tmp_path, 1, 21)

    validate_result_dataframe(result_path, 720)



"""Modules for creating a new association table from an equijoin between two catalogs"""

from .arguments import AssociationArguments
from .map_reduce import map_association, reduce_association
from .run_association import run
